from __future__ import annotations

import asyncio
from time import monotonic

from .base import SubclientBase
from ..schemas import BackgroundTaskStatus


class BgTasksClient(SubclientBase):
    async def get_task_status(self, job_id: str) -> BackgroundTaskStatus:
        if not job_id:
            raise ValueError("job_id must be a non-empty string")

        response = await self._client.get(self._client.path(f"/bgtasks/{job_id}"))
        return self._client.parse(response, BackgroundTaskStatus)

    async def wait_for_task(
        self,
        job_id: str,
        timeout_seconds: float | None = 300.0,
        poll_interval_seconds: float = 1.0,
    ) -> BackgroundTaskStatus:
        if not job_id:
            raise ValueError("job_id must be a non-empty string")
        if timeout_seconds is not None and timeout_seconds < 0:
            raise ValueError("timeout_seconds must be non-negative or None")
        if poll_interval_seconds <= 0:
            raise ValueError("poll_interval_seconds must be greater than 0")

        start = monotonic()

        while True:
            status = await self.get_task_status(job_id)

            if status.ready:
                if status.failed or not status.success:
                    raise RuntimeError(f"Background task {job_id} failed")
                return status

            if timeout_seconds is not None and monotonic() - start >= timeout_seconds:
                raise TimeoutError(
                    f"Timed out while waiting for background task {job_id}"
                )

            await asyncio.sleep(poll_interval_seconds)



from __future__ import annotations

from urllib.parse import quote
from uuid import UUID

from .base import SubclientBase
from ..schemas import CatalogItemTypeResponse


class CatalogClient(SubclientBase):
    async def map_item_type(self, key: str) -> CatalogItemTypeResponse:
        if not key:
            raise ValueError("Path key must be a non-empty string")
        response = await self._client.get(
            self._client.path(f"/catalog/item-types/path/{quote(key)}")
        )
        return self._client.parse(response, CatalogItemTypeResponse)

    async def map_item_type_id(self, key: str) -> UUID:
        response = await self.map_item_type(key)
        return response.item_type_id



from __future__ import annotations

from uuid import UUID

from .base import SubclientBase
from ..schemas import EventResponse


class EventsClient(SubclientBase):

    async def list_events(
        self,
        event_types: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[EventResponse]:
        headers: dict[str, str] = {}
        if event_types:
            headers["event-types"] = ",".join(event_types)
        if limit is not None:
            headers["limit"] = str(limit)
        if offset is not None:
            headers["offset"] = str(offset)

        response = await self._client.get_with_headers(
            self._client.path("/events"),
            headers=headers or None,
        )
        return self._client.parse_list(response, EventResponse)

    async def list_aggregate_events(
        self,
        aggregate_id: UUID,
        version: str | None = None,
    ) -> list[EventResponse]:
        params = {"version": version} if version is not None else None
        response = await self._client.get(
            self._client.path(f"/events/aggregate/{aggregate_id}/history"),
            params=params,
        )
        return self._client.parse_list(response, EventResponse)

    async def query_item_events(
        self,
        item_id: UUID,
        version: str | None = None,
    ) -> list[EventResponse]:
        params = {"version": version} if version is not None else None
        response = await self._client.get(
            self._client.path(f"/events/items/{item_id}"),
            params=params,
        )
        return self._client.parse_list(response, EventResponse)

    async def query_item_last_event(
        self,
        item_id: UUID,
        version: str | None = None,
    ) -> EventResponse:
        params = {"version": version} if version is not None else None
        response = await self._client.get(
            self._client.path(f"/events/items/{item_id}/last"),
            params=params,
        )
        return self._client.parse(response, EventResponse)

    async def query_item_last_n_events(
        self,
        item_id: UUID,
        n: int,
        version: str | None = None,
    ) -> list[EventResponse]:
        params = {"version": version} if version is not None else None
        response = await self._client.get(
            self._client.path(f"/events/items/{item_id}/last/{n}"),
            params=params,
        )
        return self._client.parse_list(response, EventResponse)



from __future__ import annotations

from datetime import date, datetime
from enum import StrEnum
from typing import Any
from uuid import UUID

from .model import AliasChoices, Field, Schema


class CommandPayload(Schema):
    aggregate_id: UUID | None = Field(default=None)
    command_type: str
    command_data: dict[str, Any] = Field(default_factory=dict)


class CreateTaskRequest(Schema):
    cr_ref: str
    tags: list[str] = Field(default_factory=list)
    message: str
    commands: list[CommandPayload]


class TaskResponse(Schema):
    id: int
    task_id: UUID
    created_by: UUID
    created_on: datetime
    resolved_by: UUID | None
    resolved_on: datetime | None
    last_updated_on: datetime | None
    last_updated_by: UUID | None
    status: str
    change_request_ref: str
    tags: list[str] = Field(default_factory=list)
    assignee: UUID | None
    message: str
    creation_message: str


class QueryRequest(Schema):
    item_type_id: UUID
    versions: list[str] = Field(default_factory=list)
    filters: list[dict[str, Any]] = Field(default_factory=list)
    filters_mode: str = Field(default="or")
    order_by: list[dict[str, Any]] = Field(default_factory=list)
    item_id: UUID | None = None
    limit: int | None = None
    offset: int | None = None
    version: str | None = None
    valid_at: date | None = None
    include_history: bool | None = False


class QueryResponseEntry(Schema):
    id: int
    item_id: UUID = Field(
        validation_alias=AliasChoices("itemId", "aggregateId", "aggregate_id"),
        serialization_alias="itemId",
    )
    version: str
    data: dict[str, Any] = Field(default_factory=dict)
    ref_properties: dict[str, Any] = Field(default_factory=dict)
    val_from: datetime
    val_to: datetime | None
    status: str | None
    release: int
    item_type_id: UUID
    task_uid: UUID | None


class QueryResponse(Schema):
    count: int
    data: list[QueryResponseEntry] = Field(default_factory=list)


class ItemRequest(Schema):
    item_type_id: UUID
    version: str = "default"
    aggregate_id: UUID = Field(
        validation_alias=AliasChoices("itemId", "aggregateId", "aggregate_id"),
        serialization_alias="itemId",
    )


class ItemDataResponse(ItemRequest):
    data: dict[str, Any] = Field(default_factory=dict)


class DependencyResponse(ItemDataResponse):
    step: int


class CatalogItemTypeResponse(Schema):
    item_type_id: UUID
    path: str


class ItemTypeVersionMetadataResponse(Schema):
    version: str
    description: str | None = None
    json_schema: dict[str, Any] = Field(default_factory=dict)


class ItemTypeMetadataResponse(Schema):
    name: str
    title: str | None = None
    ordinal: int | None = None
    collection_id: UUID
    item_type_id: UUID
    versions: list[ItemTypeVersionMetadataResponse] = Field(default_factory=list)


class ItemTypeDisplayAttributesResponse(Schema):
    item_type_id: UUID
    version: str
    display_attributes: list[str] = Field(default_factory=list)


class CommandTypes(StrEnum):
    ADD = "add"
    CREATE = "create"
    UPDATE = "update"
    RETIRE = "retire"
    REFRESH = "refresh"
    RETIRE_BULK = "retire_bulk"


class CommandDryRunRequest(Schema):
    command_uid: UUID | None = None
    item_id: UUID | None = None
    item_type_id: UUID
    version: str | None = None
    command_type: CommandTypes
    data: dict[str, Any] = Field(default_factory=dict)
    skip_nested_commands: bool = False


class ItemProjectionEventSchema(Schema):
    event_uid: UUID
    aggregate_id: UUID
    release: int | None = None
    task_uid: UUID | None = None
    event_type: str | None = None
    item_type_id: UUID | None = None
    version: str | None = None
    data: dict[str, Any] | None = None
    qfilter: list[Any] | dict[str, Any] | None = None


class CommandDryRunSchema(Schema):
    events: list[ItemProjectionEventSchema] = Field(default_factory=list)
    skip_events: list[ItemProjectionEventSchema] | None = None
    validation_errors: list[dict[str, Any]] | None = None
    graph: dict[str, Any] | None = None
    timestamp: datetime | None = None


class CommandDryRunResponseEntry(Schema):
    command_uid: UUID = Field(
        validation_alias=AliasChoices("commandUid", "commandId", "command_uid"),
        serialization_alias="commandUid",
    )
    dryrun: CommandDryRunSchema | None
    error: list[dict[str, Any]] | None


class CommandDryRunResponse(Schema):
    skip_nested_commands: bool = False
    commands: list[CommandDryRunResponseEntry] = Field(default_factory=list)


class ReportGenerationResponse(Schema):
    job_id: str


class ReportDownloadResponse(Schema):
    job_id: str
    file_name: str
    download_url: str


class BackgroundTaskStatus(Schema):
    job_id: str
    ready: bool
    failed: bool
    success: bool


class EventResponse(Schema):
    event_uid: UUID
    timestamp: datetime | None = None
    release: int | None = None
    event_type: str
    data: dict[str, Any] = Field(default_factory=dict)
    aggregate_id: UUID
    user_uid: UUID
    task_uid: UUID | None = None


__all__ = [
    "BackgroundTaskStatus",
    "CatalogItemTypeResponse",
    "CommandDryRunRequest",
    "CommandDryRunResponse",
    "CommandDryRunResponseEntry",
    "CommandDryRunSchema",
    "CommandPayload",
    "CommandTypes",
    "CreateTaskRequest",
    "DependencyResponse",
    "EventResponse",
    "ItemTypeDisplayAttributesResponse",
    "ItemTypeMetadataResponse",
    "ItemDataResponse",
    "ItemProjectionEventSchema",
    "ItemRequest",
    "ItemTypeVersionMetadataResponse",
    "QueryRequest",
    "QueryResponse",
    "QueryResponseEntry",
    "ReportDownloadResponse",
    "ReportGenerationResponse",
    "TaskResponse",
]



"""Async facade client for DSG Core API."""

from __future__ import annotations

from datetime import date
from typing import Any, TypeVar, cast
from uuid import UUID

from httpx import HTTPStatusError
from pydantic import BaseModel

from datasdr_base_api_client import BaseHTTPClient, JSON_TYPE, TokenProvider

from .item_type_cache import ItemTypePathCache
from .schemas import (
    CatalogItemTypeResponse,
    CommandDryRunRequest,
    CommandDryRunResponse,
    CreateTaskRequest,
    DependencyResponse,
    EventResponse,
    ItemDataResponse,
    ItemTypeDisplayAttributesResponse,
    ItemTypeMetadataResponse,
    ItemRequest,
    QueryRequest,
    QueryResponse,
    ReportGenerationResponse,
    TaskResponse,
)
from .subclients import (
    BgTasksClient,
    CatalogClient,
    CommandsClient,
    EventsClient,
    MetadataClient,
    QueriesClient,
    ReportsClient,
    TasksClient,
)

SchemaT = TypeVar("SchemaT", bound=BaseModel)


class DSGCoreClient(BaseHTTPClient):
    """Typed async facade client for DSG Core.

    Access grouped subclients through:
    - `client.tasks`
    - `client.bgtasks`
    - `client.queries`
    - `client.events`
    - `client.catalog`
    - `client.metadata`
    - `client.commands`
    - `client.reports`
    - `client.mapper`
    """

    def __init__(
        self,
        base_url: str,
        token: str | None = None,
        timeout: int = 30,
        token_provider: TokenProvider | None = None,
        auth_retry_attempts: int = 3,
        api_prefix: str = "/v1",
    ):
        super().__init__(
            base_url=base_url,
            token=token,
            timeout=timeout,
            token_provider=token_provider,
            auth_retry_attempts=auth_retry_attempts,
        )
        self.api_prefix = api_prefix.rstrip("/")

        self.tasks = TasksClient(self)
        self.bgtasks = BgTasksClient(self)
        self.queries = QueriesClient(self)
        self.events = EventsClient(self)
        self.catalog = CatalogClient(self)
        self.metadata = MetadataClient(self)
        self.commands = CommandsClient(self)
        self.reports = ReportsClient(self)
        self.item_type_cache = ItemTypePathCache.get_instance(self.catalog)
        self.mapper = self.item_type_cache

    def _path(self, endpoint: str) -> str:
        endpoint_path = endpoint if endpoint.startswith("/") else f"/{endpoint}"
        return f"{self.api_prefix}{endpoint_path}"

    def path(self, endpoint: str) -> str:
        return self._path(endpoint)

    @staticmethod
    def _dump(model: BaseModel) -> dict[str, Any]:
        return model.model_dump(mode="json", by_alias=True, exclude_none=True)

    @staticmethod
    def dump(model: BaseModel) -> dict[str, Any]:
        return DSGCoreClient._dump(model)

    def _parse(self, payload: JSON_TYPE, schema: type[SchemaT]) -> SchemaT:
        if payload is None:
            raise ValueError(f"Expected {schema.__name__} payload but got no content")
        return schema.model_validate(payload)

    def parse(self, payload: JSON_TYPE, schema: type[SchemaT]) -> SchemaT:
        return self._parse(payload, schema)

    def _parse_list(self, payload: JSON_TYPE, schema: type[SchemaT]) -> list[SchemaT]:
        if not payload:
            return []
        if not isinstance(payload, list):
            raise ValueError(f"Expected list payload for {schema.__name__}")
        return [schema.model_validate(item) for item in payload]

    def parse_list(self, payload: JSON_TYPE, schema: type[SchemaT]) -> list[SchemaT]:
        return self._parse_list(payload, schema)

    async def get_bytes(
        self, endpoint: str, params: dict[str, Any] | None = None
    ) -> bytes:
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        for attempt in range(1, self.auth_retry_attempts + 1):
            headers = await self._get_headers()
            response = await self.client.request(
                "GET",
                url,
                headers=headers,
                follow_redirects=self.follow_redirects,
                params=params,
            )
            try:
                response.raise_for_status()
                return response.content
            except HTTPStatusError as exc:
                status_code = exc.response.status_code
                should_retry = (
                    self.token_provider is not None
                    and status_code in {401, 403}
                    and attempt < self.auth_retry_attempts
                )
                if should_retry:
                    if self.token_provider is not None:
                        self.token_provider.invalidate_token()
                    continue
                raise
        return b""

    async def get_with_headers(
        self,
        endpoint: str,
        *,
        params: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
    ) -> JSON_TYPE:
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        for attempt in range(1, self.auth_retry_attempts + 1):
            request_headers = await self._get_headers(headers)
            response = await self.client.request(
                "GET",
                url,
                headers=request_headers,
                follow_redirects=self.follow_redirects,
                params=params,
            )
            try:
                response.raise_for_status()
                content_type = response.headers.get("content-type", "")
                if "application/json" in content_type:
                    return cast(JSON_TYPE, response.json())
                return cast(JSON_TYPE, response.text)
            except HTTPStatusError as exc:
                status_code = exc.response.status_code
                should_retry = (
                    self.token_provider is not None
                    and status_code in {401, 403}
                    and attempt < self.auth_retry_attempts
                )
                if should_retry:
                    if self.token_provider is not None:
                        self.token_provider.invalidate_token()
                    continue
                raise
        return None

    async def create_task(self, request: CreateTaskRequest) -> TaskResponse:
        return await self.tasks.create_task(request)

    async def list_tasks(
        self,
        created_by: UUID | None = None,
        assignee: UUID | None = None,
        offset: int = 0,
        limit: int = 1000,
        item_id: UUID | None = None,
    ) -> list[TaskResponse]:
        return await self.tasks.list_tasks(
            created_by=created_by,
            assignee=assignee,
            offset=offset,
            limit=limit,
            item_id=item_id,
        )

    async def get_task(self, task_id: UUID) -> TaskResponse:
        return await self.tasks.get_task(task_id)

    async def query_items(self, request: QueryRequest) -> QueryResponse:
        return await self.queries.query_items(request)

    async def query_item_dependencies(
        self,
        request: ItemRequest,
        steps: int = 1,
        limit: int | None = None,
    ) -> list[DependencyResponse]:
        return await self.queries.query_item_dependencies(
            request=request,
            steps=steps,
            limit=limit,
        )

    async def query_item_references(
        self, request: ItemRequest
    ) -> list[ItemDataResponse]:
        return await self.queries.query_item_references(request)

    async def query_item_graph(
        self,
        request: ItemRequest,
        steps: int = 3,
        data_limit: int | None = None,
        data_limit_mode: str = "total_per_item_type",
    ) -> dict[str, Any]:
        return await self.queries.query_item_graph(
            request=request,
            steps=steps,
            data_limit=data_limit,
            data_limit_mode=data_limit_mode,
        )

    async def list_events(
        self,
        event_types: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[EventResponse]:
        return await self.events.list_events(
            event_types=event_types,
            limit=limit,
            offset=offset,
        )

    async def list_aggregate_events(
        self,
        aggregate_id: UUID,
        version: str | None = None,
    ) -> list[EventResponse]:
        return await self.events.list_aggregate_events(
            aggregate_id=aggregate_id,
            version=version,
        )

    async def query_item_events(
        self,
        item_id: UUID,
        version: str | None = None,
    ) -> list[EventResponse]:
        return await self.events.query_item_events(item_id, version=version)

    async def query_item_last_event(
        self,
        item_id: UUID,
        version: str | None = None,
    ) -> EventResponse:
        return await self.events.query_item_last_event(item_id, version=version)

    async def query_item_last_n_events(
        self,
        item_id: UUID,
        n: int,
        version: str | None = None,
    ) -> list[EventResponse]:
        return await self.events.query_item_last_n_events(
            item_id,
            n,
            version=version,
        )

    async def map_item_type(self, key: str) -> CatalogItemTypeResponse:
        return await self.catalog.map_item_type(key)

    async def map_item_type_id(self, key: str) -> UUID:
        return await self.catalog.map_item_type_id(key)

    async def get_item_type_metadata(
        self,
        item_type_id: UUID,
    ) -> ItemTypeMetadataResponse:
        return await self.metadata.get_item_type_metadata(item_type_id)

    async def get_item_type_display_attributes(
        self,
        item_type_id: UUID,
        version: str,
    ) -> ItemTypeDisplayAttributesResponse:
        return await self.metadata.get_item_type_display_attributes(
            item_type_id=item_type_id,
            version=version,
        )

    async def start_dryrun_commands(
        self, commands: list[CommandDryRunRequest]
    ) -> ReportGenerationResponse:
        return await self.commands.start_dryrun_commands(commands)

    async def get_dryrun_commands_result(
        self,
        job_id: str,
        timeout_seconds: float | None = 300.0,
        poll_interval_seconds: float = 1.0,
        wait_for_ready: bool = True,
    ) -> CommandDryRunResponse:
        return await self.commands.get_dryrun_commands_result(
            job_id=job_id,
            timeout_seconds=timeout_seconds,
            poll_interval_seconds=poll_interval_seconds,
            wait_for_ready=wait_for_ready,
        )

    async def dryrun_commands(
        self,
        commands: list[CommandDryRunRequest],
        timeout_seconds: float | None = 300.0,
        poll_interval_seconds: float = 1.0,
        wait_for_ready: bool = True,
    ) -> CommandDryRunResponse:
        return await self.commands.dryrun_commands(
            commands=commands,
            timeout_seconds=timeout_seconds,
            poll_interval_seconds=poll_interval_seconds,
            wait_for_ready=wait_for_ready,
        )

    async def query(
        self,
        item_type_id: UUID,
        item_id: UUID | None = None,
        versions: list[str] | None = None,
        valid_at: date | str | None = None,
        include_history: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        filters: list[dict[str, Any]] | None = None,
        order_by: list[dict[str, Any]] | None = None,
    ) -> QueryResponse:
        valid_at_arg = valid_at.isoformat() if isinstance(valid_at, date) else valid_at
        return await self.queries.query(
            item_type_id=item_type_id,
            item_id=item_id,
            versions=versions,
            valid_at=valid_at_arg,
            include_history=include_history,
            limit=limit,
            offset=offset,
            filters=filters,
            order_by=order_by,
        )

    async def get_item_dependencies(
        self, item_type_id: UUID, item_id: UUID, version: str
    ) -> list[DependencyResponse]:
        return await self.queries.get_item_dependencies(item_type_id, item_id, version)

    async def get_item_graph(
        self,
        item_type_id: UUID,
        item_id: UUID,
        version: str,
        steps: int = 3,
        data_limit: int | None = None,
        data_limit_mode: str = "total_per_item_type",
    ) -> dict[str, Any]:
        return await self.queries.get_item_graph(
            item_type_id=item_type_id,
            item_id=item_id,
            version=version,
            steps=steps,
            data_limit=data_limit,
            data_limit_mode=data_limit_mode,
        )

    async def commands_dryrun(
        self,
        commands: list[CommandDryRunRequest],
        timeout_seconds: float | None = 300.0,
        poll_interval_seconds: float = 1.0,
        wait_for_ready: bool = True,
    ) -> CommandDryRunResponse:
        return await self.commands.commands_dryrun(
            commands=commands,
            timeout_seconds=timeout_seconds,
            poll_interval_seconds=poll_interval_seconds,
            wait_for_ready=wait_for_ready,
        )

    async def start_report_generation(
        self,
        request: QueryRequest,
        include_alias: bool = False,
        output_format: str = "excel",
    ) -> ReportGenerationResponse:
        return await self.reports.start_report_generation(
            request=request, include_alias=include_alias, output_format=output_format
        )

    async def download_report(
        self,
        job_id: str,
        timeout_seconds: float | None = 300.0,
        poll_interval_seconds: float = 1.0,
        wait_for_ready: bool = True,
    ) -> bytes:
        return await self.reports.download_report(
            job_id=job_id,
            timeout_seconds=timeout_seconds,
            poll_interval_seconds=poll_interval_seconds,
            wait_for_ready=wait_for_ready,
        )



from uuid import UUID
from unittest.mock import AsyncMock

import pytest

from dsg_core_client import DSGCoreClient
from dsg_core_client.schemas import CommandDryRunRequest, CommandTypes


@pytest.mark.asyncio
async def test_start_dryrun_commands_calls_job_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        client.post = AsyncMock(return_value={"jobId": "job-123"})  # type: ignore[method-assign]

        commands = [
            CommandDryRunRequest(
                item_type_id=UUID("00000000-0000-0000-0000-000000000001"),
                version="1.0",
                command_type=CommandTypes.CREATE,
                data={"foo": "bar"},
            )
        ]

        response = await client.commands.start_dryrun_commands(commands)

        assert response.job_id == "job-123"
        client.post.assert_awaited_once_with(
            client.path("/commands/data/dry-run"),
            json=[client.dump(command) for command in commands],
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_get_dryrun_commands_result_waits_and_fetches() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        client.get = AsyncMock(return_value={"commands": []})  # type: ignore[method-assign]
        client.bgtasks.wait_for_task = AsyncMock()  # type: ignore[method-assign]

        response = await client.commands.get_dryrun_commands_result("job-456")

        assert response.commands == []
        client.bgtasks.wait_for_task.assert_awaited_once_with(
            job_id="job-456",
            timeout_seconds=300.0,
            poll_interval_seconds=1.0,
        )
        client.get.assert_awaited_once_with(
            client.path("/commands/data/dry-run/job-456")
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_dryrun_commands_starts_job_then_fetches_result() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        client.post = AsyncMock(return_value={"jobId": "job-789"})  # type: ignore[method-assign]
        client.get = AsyncMock(return_value={"commands": []})  # type: ignore[method-assign]
        client.bgtasks.wait_for_task = AsyncMock()  # type: ignore[method-assign]

        commands = [
            CommandDryRunRequest(
                item_type_id=UUID("00000000-0000-0000-0000-000000000001"),
                version="1.0",
                command_type=CommandTypes.UPDATE,
                data={"foo": "baz"},
            )
        ]

        response = await client.commands.dryrun_commands(commands)

        assert response.commands == []
        client.post.assert_awaited_once_with(
            client.path("/commands/data/dry-run"),
            json=[client.dump(command) for command in commands],
        )
        client.bgtasks.wait_for_task.assert_awaited_once_with(
            job_id="job-789",
            timeout_seconds=300.0,
            poll_interval_seconds=1.0,
        )
        client.get.assert_awaited_once_with(
            client.path("/commands/data/dry-run/job-789")
        )
    finally:
        await client.close()



from uuid import UUID
from unittest.mock import AsyncMock

import pytest

from dsg_core_client import DSGCoreClient


@pytest.mark.asyncio
async def test_get_item_type_metadata_calls_queries_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_type_id = UUID("00000000-0000-0000-0000-000000000001")
        client.get = AsyncMock(
            return_value={
                "name": "MyType",
                "title": "My Type",
                "ordinal": 1,
                "collectionId": "00000000-0000-0000-0000-000000000002",
                "itemTypeId": str(item_type_id),
                "versions": [
                    {
                        "version": "1.0",
                        "description": "Initial version",
                        "jsonSchema": {"properties": {}},
                    }
                ],
            }
        )  # type: ignore[method-assign]

        response = await client.metadata.get_item_type_metadata(item_type_id)

        assert response.item_type_id == item_type_id
        assert response.versions[0].version == "1.0"
        client.get.assert_awaited_once_with(
            client.path(f"/queries/item-types/{item_type_id}")
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_get_item_type_display_attributes_calls_metadata_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_type_id = UUID("00000000-0000-0000-0000-000000000001")
        client.get = AsyncMock(
            return_value={
                "itemTypeId": str(item_type_id),
                "version": "1.0",
                "displayAttributes": ["name", "code"],
            }
        )  # type: ignore[method-assign]

        response = await client.metadata.get_item_type_display_attributes(
            item_type_id=item_type_id,
            version="1.0",
        )

        assert response.item_type_id == item_type_id
        assert response.display_attributes == ["name", "code"]
        client.get.assert_awaited_once_with(
            client.path(f"/metadata/json-schema/{item_type_id}/display-attributes"),
            params={"version": "1.0"},
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_get_item_type_display_attributes_rejects_empty_version() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_type_id = UUID("00000000-0000-0000-0000-000000000001")

        with pytest.raises(ValueError, match="version must be a non-empty string"):
            await client.metadata.get_item_type_display_attributes(
                item_type_id=item_type_id,
                version="",
            )
    finally:
        await client.close()



from .catalog import CatalogClient
from .bgtasks import BgTasksClient
from .commands import CommandsClient
from .events import EventsClient
from .metadata import MetadataClient
from .queries import QueriesClient
from .reports import ReportsClient
from .tasks import TasksClient

__all__ = [
    "BgTasksClient",
    "TasksClient",
    "QueriesClient",
    "CatalogClient",
    "MetadataClient",
    "CommandsClient",
    "EventsClient",
    "ReportsClient",
]



from unittest.mock import AsyncMock

import pytest

from dsg_core_client import DSGCoreClient
from dsg_core_client.schemas import BackgroundTaskStatus


@pytest.mark.asyncio
async def test_get_task_status_calls_bgtasks_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        get_mock = AsyncMock(
            return_value={
                "jobId": "job-1",
                "ready": True,
                "success": True,
                "failed": False,
            }
        )
        setattr(client, "get", get_mock)

        response = await client.bgtasks.get_task_status("job-1")

        assert response == BackgroundTaskStatus(
            job_id="job-1", ready=True, success=True, failed=False
        )
        get_mock.assert_awaited_once_with(client.path("/bgtasks/job-1"))
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_wait_for_task_returns_when_ready_success() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        get_task_status_mock = AsyncMock(
            side_effect=[
                BackgroundTaskStatus(
                    job_id="job-2", ready=False, success=False, failed=False
                ),
                BackgroundTaskStatus(
                    job_id="job-2", ready=True, success=True, failed=False
                ),
            ]
        )
        setattr(client.bgtasks, "get_task_status", get_task_status_mock)

        response = await client.bgtasks.wait_for_task(
            "job-2", timeout_seconds=1.0, poll_interval_seconds=0.001
        )

        assert response.job_id == "job-2"
        assert response.ready is True
        assert response.success is True
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_wait_for_task_raises_for_failed_task() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        get_task_status_mock = AsyncMock(
            return_value=BackgroundTaskStatus(
                job_id="job-3", ready=True, success=False, failed=True
            )
        )
        setattr(client.bgtasks, "get_task_status", get_task_status_mock)

        with pytest.raises(RuntimeError, match="Background task job-3 failed"):
            await client.bgtasks.wait_for_task("job-3")
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_wait_for_task_raises_timeout() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        get_task_status_mock = AsyncMock(
            return_value=BackgroundTaskStatus(
                job_id="job-4", ready=False, success=False, failed=False
            )
        )
        setattr(client.bgtasks, "get_task_status", get_task_status_mock)

        with pytest.raises(TimeoutError, match="Timed out while waiting"):
            await client.bgtasks.wait_for_task(
                "job-4", timeout_seconds=0, poll_interval_seconds=0.001
            )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_wait_for_task_rejects_invalid_poll_interval() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        with pytest.raises(
            ValueError, match="poll_interval_seconds must be greater than 0"
        ):
            await client.bgtasks.wait_for_task("job-5", poll_interval_seconds=0)
    finally:
        await client.close()



from src.dsg_core.utils.mappings import deep_update


def test_deep_update():
    dict1 = {
        "properties": {
            "name": {
                "type": "string",
                "minLength": 1,
            }
        },
        "required": ["name"],
    }

    dict2 = {
        "properties": {
            "name": {
                "maxLength": 255,
                "minLength": 3,
            }
        },
    }
    result = deep_update(dict1, dict2)
    print(result)
    assert result == {
        "properties": {
            "name": {
                "type": "string",
                "maxLength": 255,
                "minLength": 3,
            }
        },
        "required": ["name"],
    }



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import DSGCoreClient


class SubclientBase:
    def __init__(self, client: "DSGCoreClient"):
        self._client = client



from types import SimpleNamespace
from unittest.mock import MagicMock
from uuid import uuid4

from dsg_core.services.data.commandDryrunService import CommandDryrunService
from dsg_core.services.data.dependencies import RefsAndDepsSchema
from dsg_core.services.data.dataCommandProcessor import (
    DataCommandProcessor,
    DataCommandProcessorOutput,
)


def test_data_command_processor_forwards_skip_nested_commands() -> None:
    expected_output = DataCommandProcessorOutput(events=[], projection_events=[])

    class StubHandler:
        def __init__(self, session):
            self.process = MagicMock(return_value=expected_output)

    processor = DataCommandProcessor(handler=StubHandler, session=MagicMock())
    command = MagicMock()
    user_uid = uuid4()

    output = processor.process(
        command,
        user_uid=user_uid,
        skip_nested_commands=True,
    )

    assert output is expected_output
    processor.handler.process.assert_called_once_with(
        command,
        user_uid=user_uid,
        task_id=None,
        validate_only=False,
        skip_nested_commands=True,
    )


def test_dryrun_service_propagates_skip_nested_commands(monkeypatch) -> None:
    processor_calls: list[bool] = []

    class StubProcessor:
        def __init__(self, handler, session):
            self.handler = handler
            self.session = session

        def process(self, command, *, user_uid, skip_nested_commands=False, **kwargs):
            processor_calls.append(skip_nested_commands)
            return DataCommandProcessorOutput(events=[], projection_events=[])

    class StubDryrun:
        def __init__(self) -> None:
            self.graph = None

        def model_dump(self, mode="json"):
            return {
                "events": [],
                "skipEvents": None,
                "validationErrors": None,
                "graph": self.graph,
                "timestamp": None,
            }

    monkeypatch.setattr(
        "dsg_core.services.data.commandDryrunService.DataCommandProcessor",
        StubProcessor,
    )
    monkeypatch.setattr(
        "dsg_core.services.data.commandDryrunService.DependencyService.retrieve_item_references_and_dependencies_from_dryrun",
        lambda *args, **kwargs: RefsAndDepsSchema(references=[], dependencies=[]),
    )
    monkeypatch.setattr(
        "dsg_core.services.data.commandDryrunService.TaskDryrunService.task_dryrun_schema_from_command_output",
        lambda **kwargs: StubDryrun(),
    )
    monkeypatch.setattr(
        "dsg_core.services.data.commandDryrunService.TaskDryrunService.compile_command_dryrun_graph",
        lambda **kwargs: None,
    )

    session = MagicMock()
    response = CommandDryrunService.dryrun_commands(
        payload=[
            {
                "command_uid": uuid4(),
                "item_id": uuid4(),
                "item_type_id": uuid4(),
                "version": "1.0",
                "command_type": "create",
                "data": {},
                "skip_nested_commands": True,
            }
        ],
        user_uid=uuid4(),
        session=session,
    )

    assert processor_calls == [True]
    assert response.skip_nested_commands is True
    session.rollback.assert_called_once_with()



from __future__ import annotations

from uuid import UUID

from .base import SubclientBase
from ..schemas import ItemTypeDisplayAttributesResponse, ItemTypeMetadataResponse


class MetadataClient(SubclientBase):
    async def get_item_type_metadata(
        self,
        item_type_id: UUID,
    ) -> ItemTypeMetadataResponse:
        response = await self._client.get(
            self._client.path(f"/queries/item-types/{item_type_id}")
        )
        return self._client.parse(response, ItemTypeMetadataResponse)

    async def get_item_type_display_attributes(
        self,
        item_type_id: UUID,
        version: str,
    ) -> ItemTypeDisplayAttributesResponse:
        if not version:
            raise ValueError("version must be a non-empty string")

        response = await self._client.get(
            self._client.path(
                f"/metadata/json-schema/{item_type_id}/display-attributes"
            ),
            params={"version": version},
        )
        return self._client.parse(response, ItemTypeDisplayAttributesResponse)



from uuid import UUID
from unittest.mock import AsyncMock

import pytest

from dsg_core_client.client import DSGCoreClient


@pytest.mark.asyncio
async def test_list_events_calls_events_collection_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        get_mock = AsyncMock(return_value=[])
        setattr(client, "get_with_headers", get_mock)

        response = await client.events.list_events(
            event_types=["ItemCreatedEvent", "ItemUpdatedEvent"],
            limit=10,
            offset=5,
        )

        assert response == []
        get_mock.assert_awaited_once_with(
            client.path("/events"),
            headers={
                "event-types": "ItemCreatedEvent,ItemUpdatedEvent",
                "limit": "10",
                "offset": "5",
            },
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_list_aggregate_events_calls_history_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        aggregate_id = UUID("00000000-0000-0000-0000-000000000001")

        get_mock = AsyncMock(return_value=[])
        setattr(client, "get", get_mock)

        response = await client.events.list_aggregate_events(
            aggregate_id,
            version="2.0",
        )

        assert response == []
        get_mock.assert_awaited_once_with(
            client.path(f"/events/aggregate/{aggregate_id}/history"),
            params={"version": "2.0"},
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_query_item_events_calls_events_item_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_id = UUID("00000000-0000-0000-0000-000000000001")
        user_id = UUID("00000000-0000-0000-0000-000000000002")

        get_mock = AsyncMock(
            return_value=[
                {
                    "eventUid": "00000000-0000-0000-0000-000000000003",
                    "timestamp": "2026-01-01T00:00:00Z",
                    "release": 1,
                    "eventType": "ItemCreatedEvent",
                    "data": {"name": "item-1"},
                    "aggregateId": str(item_id),
                    "userUid": str(user_id),
                    "taskUid": None,
                }
            ]
        )
        setattr(client, "get", get_mock)

        response = await client.events.query_item_events(item_id)

        assert len(response) == 1
        assert response[0].event_type == "ItemCreatedEvent"
        assert response[0].aggregate_id == item_id
        get_mock.assert_awaited_once_with(client.path(f"/events/items/{item_id}"))
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_query_item_events_passes_version_query_param() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_id = UUID("00000000-0000-0000-0000-000000000001")

        get_mock = AsyncMock(return_value=[])
        setattr(client, "get", get_mock)

        _ = await client.events.query_item_events(item_id, version="2.0")

        get_mock.assert_awaited_once_with(
            client.path(f"/events/items/{item_id}"),
            params={"version": "2.0"},
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_query_item_last_event_calls_events_last_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_id = UUID("00000000-0000-0000-0000-000000000001")
        user_id = UUID("00000000-0000-0000-0000-000000000002")

        get_mock = AsyncMock(
            return_value={
                "eventUid": "00000000-0000-0000-0000-000000000003",
                "timestamp": "2026-01-01T00:00:00Z",
                "release": 2,
                "eventType": "ItemUpdatedEvent",
                "data": {"name": "item-2"},
                "aggregateId": str(item_id),
                "userUid": str(user_id),
                "taskUid": None,
            }
        )
        setattr(client, "get", get_mock)

        response = await client.events.query_item_last_event(item_id)

        assert response.event_type == "ItemUpdatedEvent"
        assert response.release == 2
        get_mock.assert_awaited_once_with(
            client.path(f"/events/items/{item_id}/last"),
            params=None,
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_query_item_last_event_passes_version_query_param() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_id = UUID("00000000-0000-0000-0000-000000000001")

        get_mock = AsyncMock(
            return_value={
                "eventUid": "00000000-0000-0000-0000-000000000003",
                "timestamp": "2026-01-01T00:00:00Z",
                "release": 2,
                "eventType": "ItemUpdatedEvent",
                "data": {"version": "2.0"},
                "aggregateId": str(item_id),
                "userUid": "00000000-0000-0000-0000-000000000002",
                "taskUid": None,
            }
        )
        setattr(client, "get", get_mock)

        _ = await client.events.query_item_last_event(item_id, version="2.0")

        get_mock.assert_awaited_once_with(
            client.path(f"/events/items/{item_id}/last"),
            params={"version": "2.0"},
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_query_item_last_n_events_calls_events_last_n_endpoint() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_id = UUID("00000000-0000-0000-0000-000000000001")
        user_id = UUID("00000000-0000-0000-0000-000000000002")

        get_mock = AsyncMock(
            return_value=[
                {
                    "eventUid": "00000000-0000-0000-0000-000000000003",
                    "timestamp": "2026-01-01T00:00:00Z",
                    "release": 3,
                    "eventType": "ItemRetiredEvent",
                    "data": {},
                    "aggregateId": str(item_id),
                    "userUid": str(user_id),
                    "taskUid": None,
                }
            ]
        )
        setattr(client, "get", get_mock)

        response = await client.events.query_item_last_n_events(item_id, 5)

        assert len(response) == 1
        assert response[0].release == 3
        get_mock.assert_awaited_once_with(
            client.path(f"/events/items/{item_id}/last/5"),
            params=None,
        )
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_query_item_last_n_events_passes_version_query_param() -> None:
    client = DSGCoreClient(base_url="https://example.org", token="token")
    try:
        item_id = UUID("00000000-0000-0000-0000-000000000001")

        get_mock = AsyncMock(return_value=[])
        setattr(client, "get", get_mock)

        _ = await client.events.query_item_last_n_events(item_id, 5, version="2.0")

        get_mock.assert_awaited_once_with(
            client.path(f"/events/items/{item_id}/last/5"),
            params={"version": "2.0"},
        )
    finally:
        await client.close()
